class ReentrantLock {
    //@ predicate ReentrantLock(predicate() inv);
    
    ReentrantLock()
        //@ requires exists<predicate()>(?inv) &*& inv();
        //@ ensures ReentrantLock(inv);
    { /*@ assume(false); @*/ }
    
    void lock()
        //@ requires [_]ReentrantLock(?inv);
        //@ ensures inv();
        // Soundness: Assumes that this thread does not yet hold this lock! Unsound in the presence of re-entry.
    { /*@ assume(false); @*/ }
    
    void unlock()
        //@ requires [_]ReentrantLock(?inv) &*& inv();
        //@ ensures true;
    { /*@ assume(false); @*/ }
    
}

class SV {
    int value;
    ReentrantLock lock;
}

final class SVList {
    //@ predicate List(list<SV> elems);
    public SVList()
        //@ requires true;
        //@ ensures List(nil);
    { /*@ assume(false); @*/ }
    public int size()
        //@ requires [?f]List(?es);
        //@ ensures [f]List(es) &*& result == length(es);
    { throw new RuntimeException(); }
    public SV get(int index)
        //@ requires [?f]List(?es) &*& 0 <= index &*& index < length(es);
        //@ ensures [f]List(es) &*& result == nth(index, es);
    { throw new RuntimeException(); }
    boolean add(SV e)
        //@ requires List(?es);
        //@ ensures List(append(es, cons(e, nil))) &*& result;
    { throw new RuntimeException(); }
}

/*@
predicate_ctor lock_inv(SV v)(;) = v.value |-> _;
predicate locks(list<SV> vs;) =
    vs == nil ?
        true
    :
        [_]head(vs).lock |-> ?lock &*& [_]lock.ReentrantLock(lock_inv(head(vs))) &*& locks(tail(vs));
                //[_]SV_lock(head(vs), ?lock) &*& [_]lock.ReentrantLock(lock_inv(head(vs))) &*& locks(tail(vs));
        
        
predicate invariants(list<SV> vs;) =
    vs == nil ?
        true
    :
        lock_inv(head(vs))() &*& invariants(tail(vs));
@*/

class Stmt {
    SVList vsList;
    /*@
    predicate Stmt_lock(list<SV> vs) =
        this.vsList |-> ?a &*& a.List(vs) &*& locks(vs);
    @*/
    void lock()
        //@ requires [_]Stmt_lock(?vs);
        //@ ensures invariants(vs);
    {
        for (int i = 0; i < vsList.size(); i++)
            //@ requires [_]vsList |-> ?b &*& [_]b.List(vs) &*& [_]locks(drop(i, vs)) &*& i >= 0 &*& i <= length(vs);
            //@ ensures invariants(drop(old_i, vs));
        {
            //@ drop_n_plus_one(i, vs);
            vsList.get(i).lock.lock();
        }
    }
    void unlock()
        //@ requires [_]Stmt_lock(?vs) &*& invariants(vs);
        //@ ensures true;
    {
        for (int i = 0; i < vsList.size(); i++)
            //@ requires [_]vsList |-> ?b &*& [_]b.List(vs) &*& [_]locks(drop(i, vs)) &*& invariants(drop(i, vs)) &*& i >= 0;
            //@ ensures true;
        {
            //@ drop_n_plus_one(i, vs);
            vsList.get(i).lock.unlock();
        }
    }
}
